#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include <quat.h>
#include <vrpn_Tracker.h>
#include "graphics.h"

#define INCHES_TO_METERS(inches) ( (inches) * 0.0254 )

int winID, menuID;

/* transformations */
q_xyz_quat_type room_from_world;
q_xyz_quat_type tracker_from_room;
q_xyz_quat_type head_sensor_from_tracker;
q_xyz_quat_type hand_sensor_from_tracker;
q_xyz_quat_type head_from_head_sensor;
q_xyz_quat_type left_eye_from_head;
q_xyz_quat_type right_eye_from_head;

q_xyz_quat_type left_eye_from_head_sensor;
q_xyz_quat_type right_eye_from_head_sensor;
q_xyz_quat_type tracker_from_world;


vrpn_Tracker_Remote *tracker;

void DrawArrow()
{
    /* 
       Draw arrow of length 1.0 from the origin pointing
       in positive z direction 
    */

    static GLUquadricObj *quadric_state = gluNewQuadric();
    gluQuadricDrawStyle(quadric_state, (GLenum)GLU_FILL);
    gluQuadricNormals(quadric_state, (GLenum)GLU_SMOOTH);

    float length = 1.0;
    float body_radius = length / 50.0;
    float body_length = length * 0.95;
    float tip_radius = body_radius * 2.0;
    float tip_length = length - body_length;
    float num_sides = 8;

    /* Draw arrow body as cylinder */
    gluCylinder(quadric_state, body_radius, body_radius, body_length,
                num_sides, 1);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslatef(0, 0, body_length);

    /* Draw Arrow Tip (cone is degenerate cylinder) */
    gluCylinder(quadric_state, tip_radius, 0, tip_length,
                num_sides, 1);

    glPopMatrix();
}


void DrawCoordinateSystem()
{
    /* Draw x-, y-, and z-axes in red, green, and blue, respectively */

    /* 
       Rotate arrow from positive-z to positive-x axis by rotating
       90 degrees clockwise when looking down positive-y axis
    */
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glRotated(90.0, 0.0, 1.0, 0.0);

    /* Set color to red */
    glColor3f(1.0, 0.0, 0.0);

    /* Draw */
    DrawArrow();

    /* Restore original matrix */
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    /* 
       Rotate arrow from positive-z to positive-y axis by rotating
       90 degrees counter-clockwise when looking down positive-x axis
    */
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glRotated(-90.0, 1.0, 0.0, 0.0);

    /* Set color to green */
    glColor3f(0.0, 1.0, 0.0);

    /* Draw */
    DrawArrow();

    /* Restore original matrix */
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    /* Set color to blue */
    glColor3f(0.0, 0.0, 1.0);

    /* Draw */
    DrawArrow();

    return;
}

void DrawFloor()
{
    /* assume we are in room coordinates (eye_from_room on MODELVIEW) */

    /* draw 6 ft by 6 ft floor in the corner of the room */
    
    glColor3f(0.5, 0.5, 0.5);

    glBegin(GL_QUADS);
      glNormal3f(0.0, 0.0, 1.0);
      glVertex3f(0.0, 0.0, 0.0);
      glVertex3f(INCHES_TO_METERS(72), 0.0, 0.0);
      glVertex3f(INCHES_TO_METERS(72), INCHES_TO_METERS(72), 0.0);
      glVertex3f(0.0, INCHES_TO_METERS(72), 0.0);
    glEnd();

    return;
}

void
DrawWorld( void )
{
    q_xyz_quat_type  world_from_tracker, world_from_room;
    qogl_matrix_type world_from_tracker_matrix, world_from_room_matrix;

    /* Assumes we start with (left/right)eye_from_world on the MODELVIEW stack */

    /* Draw coordinate axes in tracker space */

    q_xyz_quat_invert(&world_from_tracker, &tracker_from_world);
    q_xyz_quat_to_ogl_matrix(world_from_tracker_matrix, &world_from_tracker);


    /* effectively make MODELVIEW eye_from_tracker */
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixd(world_from_tracker_matrix);


    /* shrink coordinate system model to 1/2 meter instead of 1 meter */
    glScaled(0.5, 0.5, 0.5);
    DrawCoordinateSystem();

    /* return MODELVIEW to eye_from_world */
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();


    /* Draw coordinate axes in room space */

    q_xyz_quat_invert(&world_from_room, &room_from_world);
    q_xyz_quat_to_ogl_matrix(world_from_room_matrix, &world_from_room);

    /* make MODELVIEW eye_from_room */
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glMultMatrixd(world_from_room_matrix);

    DrawCoordinateSystem();
    DrawFloor();

    /* return MODELVIEW to eye_from_world */
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    return;
}

void
DrawFrame( void )
{
    q_xyz_quat_type  left_eye_from_tracker, left_eye_from_world;
    q_xyz_quat_type  right_eye_from_tracker, right_eye_from_world;
    qogl_matrix_type left_eye_from_world_matrix, right_eye_from_world_matrix;


    grStartFrame( );

    /*
    ** Set up for drawing a left eye view of the world with this xform from
    ** world space to eye space.
    */

    q_xyz_quat_compose(&left_eye_from_tracker,
                       &left_eye_from_head_sensor,
                       &head_sensor_from_tracker);
    q_xyz_quat_compose(&left_eye_from_world,
                       &left_eye_from_tracker,
                       &tracker_from_world);
    q_xyz_quat_to_ogl_matrix(left_eye_from_world_matrix, &left_eye_from_world);


    grLeftView( left_eye_from_world_matrix );
    DrawWorld( );



    /* ... and now the same for the right eye */


    q_xyz_quat_compose(&right_eye_from_tracker,
                       &right_eye_from_head_sensor,
                       &head_sensor_from_tracker);
    q_xyz_quat_compose(&right_eye_from_world,
                       &right_eye_from_tracker,
                       &tracker_from_world);
    q_xyz_quat_to_ogl_matrix(right_eye_from_world_matrix, &right_eye_from_world);

    grRightView( right_eye_from_world_matrix );
    DrawWorld( );


    grEndFrame( );

    return;
}


/* 
   Make any changes to the world, including updating the
   world_from_head_sensor and world_from_hand_sensor transformation,
   as well as any modeling transformations changed due to animation or
   object manipulation.
*/
void UpdateWorld( void )
{
    /* update the tracker sensor transformations */
    tracker->mainloop();

    /* take a step in the animation */



    /* Tell glut that things have changed, so the screen needs 
       to be redrawn */
    glutPostRedisplay( );
}

void
MenuHandler( int entry )
{
    switch( entry ) {
    case 1 : /* quit */
      exit( 0 );
      break;
    } /* end switch */

    return;
}

void tracker_handler(void *user_data, vrpn_TRACKERCB tracker_data)
{
    /* update head transformation */
    if (tracker_data.sensor == 0)
    {
       q_xyz_quat_type tracker_from_head_sensor;

       q_vec_copy(tracker_from_head_sensor.xyz, tracker_data.pos);
       q_copy(tracker_from_head_sensor.quat, tracker_data.quat);
       q_xyz_quat_invert(&head_sensor_from_tracker, &tracker_from_head_sensor);
    }

    /* update hand transformation */
    else if (tracker_data.sensor == 1)
    {
       q_xyz_quat_type tracker_from_hand_sensor;

       q_vec_copy(tracker_from_hand_sensor.xyz, tracker_data.pos);
       q_copy(tracker_from_hand_sensor.quat, tracker_data.quat);
       q_xyz_quat_invert(&hand_sensor_from_tracker, &tracker_from_hand_sensor);
    }

    return;
}

void InitTracker()
{
    /* 
       You could probably make this try several times rather than giving
       up after a single try.  However, you must check to see if the connection
       was actually made.  If you allow the application to continue without a
       successful connection, you will ultimately bring up a useless graphics
       window that covers up the window of whoever is actually running.
    */

    fprintf(stderr, "Attempting to connect to vrpn server.\n");

    tracker = new vrpn_Tracker_Remote("Tracker0@nameless.cs.jhu.edu");
    if (!tracker->connectionPtr())
    {
       fprintf(stderr, "Application couldn't connect to vrpn server.  Exiting...\n");
       exit(1);
    }

    fprintf(stderr, "Connected!\n");

    tracker->register_change_handler(NULL, tracker_handler);
}

void InitTransforms()
{
    /* initialize room_from_world */

    q_vec_set(room_from_world.xyz, 0,0,0);
    q_make(room_from_world.quat, 0,0,0,0);



    /* initialize tracker_from_room */

    /* initialize tracker_from_room */

    /* 
       Room coordinates have the origin in the corner, +x towards the 
       wall with the windows, +y towards the wall with the sink, and +z
       going up.  The tracker coordinates have the origin in the tracker
       source, +x pointing down, +y towards the wall with the windows,
       and +z pointing towards the wall shared with the vision lab.
       
       The tracking source is about 84 inches off the floor, 156 inches from
       the wall containing the doors, and 11.5 inches from the wall shared
       with the vision lab.
    */

    q_xyz_quat_type z_horizontal_from_room;
    q_vec_set(z_horizontal_from_room.xyz, 0,0,0);
    q_make(z_horizontal_from_room.quat, 1,0,0, -M_PI/2.0);

    q_xyz_quat_type x_down_z_horizontal_from_z_horizontal;
    q_vec_set(x_down_z_horizontal_from_z_horizontal.xyz, 0,0,0);
    q_make(x_down_z_horizontal_from_z_horizontal.quat,
           0,0,1, M_PI/2.0);

    q_xyz_quat_type tracker_from_x_down_z_horizontal;
    q_vec_set(tracker_from_x_down_z_horizontal.xyz,
              INCHES_TO_METERS(83.5), 
              -INCHES_TO_METERS(156.0), 
              INCHES_TO_METERS(11.5));
    q_make(tracker_from_x_down_z_horizontal.quat, 0,0,0,0);
    
    q_xyz_quat_type x_down_z_horizontal_from_room;
    q_xyz_quat_compose(&x_down_z_horizontal_from_room,
                       &x_down_z_horizontal_from_z_horizontal,
                       &z_horizontal_from_room);

    q_xyz_quat_compose(&tracker_from_room,
                       &tracker_from_x_down_z_horizontal,
                       &x_down_z_horizontal_from_room);


    /* initialize head_sensor_from_tracker (eventually set from tracking data */
    q_vec_set(head_sensor_from_tracker.xyz, 0,0,0);
    q_make(head_sensor_from_tracker.quat, 0,0,0,0);

    /* initialize hand_sensor_from_tracker (eventually set from tracking data */
    q_vec_set(hand_sensor_from_tracker.xyz, 0,0,0);
    q_make(hand_sensor_from_tracker.quat, 0,0,0,0);



    /* NOTE: You have to code up the remaining transforms on your own! */

    /* initialize head_from_head_sensor */

    /* 
       With the tracking sensor mounted on the HMD, the x-axis is towards
       the front of the HMD, the y-axis is towards the right side of the
       HMD, and the z-axis is towards the floor.  This is the sensor's
       coordinate system (with the origin at the center of the sensor.

       The head_from_head_sensor transformation should transform this
       coordinate system to head coordinates, which has the origin
       between the eyes, the x-axis pointing to the right, the y-axis pointing
       upward, and the z-axis pointing towards the back of the head.

       You can assume that eyes are about 6 inches forward of the
       sensor and about 4 inches below it.
    */
    q_vec_set(head_from_head_sensor.xyz, 0,0,0);
    q_make(head_from_head_sensor.quat, 0,0,0,0);


 
    /* A typical distance between the left and right eyes is 62 mm. */

    /* initialize left_eye_from_head */
    q_vec_set(left_eye_from_head.xyz, 0,0,0);
    q_make(left_eye_from_head.quat, 0,0,0,0);

    /* initialize right_eye_from_head */
    q_vec_set(right_eye_from_head.xyz, 0,0,0);
    q_make(right_eye_from_head.quat, 0,0,0,0);


    /* now compose all the transformations that won't be changing */

    /* compose left_eye_from_head_sensor */
    q_xyz_quat_compose(&left_eye_from_head_sensor,
                       &left_eye_from_head, &head_from_head_sensor);

    /* compose right_eye_from_head_sensor */
    q_xyz_quat_compose(&right_eye_from_head_sensor,
                       &right_eye_from_head, &head_from_head_sensor);

    /* compose tracker_from_world -- assume no flying or scaling the user/world */
    q_xyz_quat_compose(&tracker_from_world,
                       &tracker_from_room, &room_from_world);
}

void InitWorld()
{
  /* Load your model here and do any preparations for your animation */

}

void
main( int argc, char *argv[] )
{
    InitTracker();  /* first grab the tracker if it's available */
    InitTransforms();
    InitWorld();

    grInit( &argc, argv, 0, 1024, 640, 480, GL_TRUE );    

    /* Establish the function that draws the frame (completely) */
    glutDisplayFunc( DrawFrame );

    /* Establish a function to call whenever the event queue is empty */
    glutIdleFunc( UpdateWorld );

    /* Use GLUT's menu functions to set up a few options and a quit feature */
    menuID = glutCreateMenu( MenuHandler );
    glutAddMenuEntry( "Quit", 1 );
    glutAttachMenu( GLUT_RIGHT_BUTTON );

    glutMainLoop( );

    /* not reached */
    return;
} /* end main */

